'''
Библиотека lite v.0.4.1.3 (файл main.py)
Протоколирование, организация главного цикла приложений
--------------------------------------------------------------------------------
История изменений в конце файла
OldBean, 25.07.2021
'''
#-------------------------------------------------------------------------------
import sys
#-------------------------------------------------------------------------------
''' Загрузка пользовательского модуля со специфическими для конкретной задачи
параметрами, режимами и т.д. Имя модуля должно быть задано в списке параметров
командной строки как-то так: python3 main.py user_module.py

По умолчанию будет делаться попытка загрузить модуль с названием sr1
(файл sr1.py) 
'''
sp = 'sr1' # Модуль sr1.py загружается по умолчанию
if len(sys.argv) > 1:
  sp = sys.argv[1].rsplit('.', 1)[0]
try:
  exec('from ' + sp + ' import *')
except ModuleNotFoundError as mnfe: # Пользовательский модуль не загружен
  print(mnfe)
  from lite import *
#-------------------------------------------------------------------------------
''' Для подключения TCP клиента в загружаемом пользовательском модуле должен
быть задан IP-адрес TCP сервера в локальной сети. Обычно это управляющий
компьютер (малинка)
'''
if 'server_addr' not in globals():  # IP TCP-сервера в локальной сети не задан:
  server_addr = ('localhost', 5238) # значит локальная работа на хосте
print(f'Адрес и порт TCP-сервера: {server_addr}')

# Запускаем отдельный поток TCP-сервера
thr = threading.Thread(target=server_thread, args=(server_addr,), daemon=True)
thr.start()
#-------------------------------------------------------------------------------
exit_flag = False # Флаг для выхода из главного цикла приложения
old_mode = 0 # Предыдущий режим (для обнаружения факта смены режима работы)

now = time.time() # Текущее время (сек)
tim.pst = now # Начало всего процесса
tim.mst = now # Начало текущего режима

''' Флаг h_flag определяет характер информации, отображаемой на консоли
приложения. Если флаг установлен, то на экране отображается информация об
оборудовании, если сброшен, то текущее состояние системы.
'''
h_flag = False
bot_flag = False # Отправлять/не отправлять важную информацию на чат

timeout = 500 # Время ожидания нажатия какой-нибудь клавиши (мс)
tt = TT(timeout) # Переводим клавиатуру в неблокирующий режим без эха

log = open('log', 'w')
log.write(time.ctime() + '\n')
log.write(u.log_head + '\n')
log_str = 'f\'' + u.log_vars + '\'' # Шапка для данных лога

kbhit_flag = False # Индикатор нажатия клавиши на клавиатуре хоста

''' В импортируемом модуле со спецификой решаемоей задачи пользователь может
определить callback-функцию crit_info(), которая будет вызываться при важных
событиях. Эта функция может, например, посылать сообщения в телеграм-чат,
на e-mail и т.п.
'''
try:
  crit_info('Начало работы')
except:
  pass
#-------------------------------------------------------------------------------
while True: # Входим в бесконечный главный цикл приложения
  try:
    sens_conv() # Запускаем все датчики на преобразование

    ''' Нажатие клавиши на клавиатуре хоста должна иметь приоритет перед
    командами удаленных клиентов. Поэтому, при нажатии клавиши на клавиатуре
    хоста, формируется запрос, который помещается в начало очереди
    '''
    if tt.kbhit(): # Нажата клавиша на клавиатуре хоста (timeout = 0.5 сек)
      ch = tt.getch()
      try: # Добавляем запрос в очередь
        que_lock.acquire()
        req_que.append(Request(hot_keys, bmsg=ch.encode()))
      except BaseException as e:
        pass
#        print(f'Что-то не так: {e}')
      finally:
        que_lock.release()

    try: # Теперь извлекаем первый запрос в очереди запросов
      que_lock.acquire()
      req = req_que.popleft()
    except IndexError as ie: # Очередь запросов пуста
      req = Request(hot_keys, None, b'') # Пустой запрос
    finally:
      que_lock.release()

    sens_update() # Обновляем показания всех датчиков в соотв. объектах

    server_msg = '' # Сообщение, которое будет отправлено клиенту
    if req.kbhit: # Введена горячая клавиши
      if req.ch == 'q': # Конец работы
        if not req.conn: # Эта команда отрабатывается только от хоста
          u.mode = 0
          exit_flag = True
      elif req.ch >= '0' and req.ch < str(len(modes)): # Введен номер режима
        u.mode = int(req.ch)
      elif req.ch == 'h': # Выдаем информацию о найденном оборудовании
        if req.conn: # Клиентам отправляется строка
          server_msg = hinfo()
        else: # Для хоста устанавливается или снимается флаг
          h_flag = not h_flag
      elif req.ch == 'k': # Выдаем информацию о горячих клавишах
        server_msg = kinfo()
      elif req.ch == 'm': # Выдаем информацию о доступных режимах работы
        server_msg = minfo()
      elif req.ch == 'u': # Выдаем пользовательские переменные
        server_msg = uinfo()
      elif req.ch == 'v': # Выдаем значения датчиков и контроллеров
        server_msg = vinfo()
      elif req.ch == 'i': # Выдаем всю текущую информацию о системе
        server_msg = info()
      else:
        try:
          server_msg = user_hk(req.ch) # Пользовательские горячие клавиши
        except:
          pass
    elif req.msg: # Отрабатываем сообщение клиента как питоновское выражение
      expr = req.msg
      eq_pos = expr.find('=') # Ищем знак равенства
      exec_flag = eq_pos > 0 and eq_pos < len(expr) - 1 # Знак есть
      # Проверим, что это одиночный знак равенства. Т.е. - присваивание
      if exec_flag:
        exec_flag = expr[eq_pos + 1] != '='
      if exec_flag:
        exec_flag = expr[eq_pos - 1] not in '><+-*/!' 
      try:
        if exec_flag: # В выражении присваивание, значит применим exec
          exec(expr)
          server_msg = 'OK'
        else: # Если просто выражение и нужно вернуть результат - применим eval
          _ = eval(expr)
          server_msg = str(_)
      except Exception as e:
        server_msg = str(e)

    try: # Проверка условий автоматической смены режимов
      if emergency(): # Аварийная ситуация!!! Пока - немедленное отключение
        mode = 0
        exit_flag = True
    except:
      pass

    if not exit_flag: # Штатная работа - проверяем условия текущего режима
      modes[u.mode].cond()
     
    if u.mode != old_mode: # Отработка смены режима
      modes[u.mode].init() # Инициализация нового режима
      server_msg = f'Смена режима: {old_mode}->{u.mode} ({modes[u.mode].name})'
      server_msg += '\n\n' + vinfo()
      old_mode = u.mode
      tim.mst = time.time()
      try:
        crit_info(server_msg)
      except:
        pass

    # выводим информацию на консоль хоста
    print('\x1b[2J') # очистим экран
    print('\x1b[H', end='') # Курсор - домой
    print()
    if h_flag: # Нужно выводить информацию об оборудовании
      print(hinfo()) 
      print('Нажмите \'h\' для выхода...')
    else: # Нужно выводить информацию о текущем состоянии системы
      info1 = info(all_keys=False).rstrip()
      l = len(info1) - info1.rfind('\n')
      info2 = f'Клиентов: {len(clients)}; запросов: {len(req_que)}'
      print(info1 + (80 - l - len(info2))*' ' + info2)

    # Пишем информацию в лог
    exec('now = time.time(); wrk = ' + log_str) # Формируем строку лога
    log.write(wrk + '\n')
    log.flush()

    if exit_flag:
      break
    else:
      time.sleep(0.1)

    # Если нужно, отправляем информацию клиенту
    if req.conn:
      req.conn.send(server_msg.encode())

    del(req)

  except KeyboardInterrupt: # Грубо завершаем работу (по ctrl+c)
    break
  except Exception as e:
    server_msg = f'\nОшибка: {e}'
    print(server_msg)
    try:
      crit_info(server_msg)
    except:
      pass
    time.sleep(10)
# Конец главного цикла приложения ----------------------------------------------
u.mode = 0 # Переходим в режим мониторинга
modes[u.mode].init() # Инициализируем его
if len(clients) > 0: # Прощаемся с подключенными клиентами
  for conn in clients:
    conn.send('bye'.encode())
try: # Рассылаем критическое сообщение о завершении работы
  crit_info('Завершение работы')
except:
  pass
del(tt) # Переводим консоль в обычный режим
log.close()
#-------------------------------------------------------------------------------
''' Список изменений
25.07.2021 1. Косметические правки, добавление комментариев
'''
#-------------------------------------------------------------------------------
